@@ -1,6 +1,6 @@
 package SVN::Mirror::Ra;
 @ISA = ('SVN::Mirror');
-$VERSION = '0.66';
+$VERSION = '0.68';
 use strict;
 use SVN::Core;
 use SVN::Repos;
@@ -413,6 +413,7 @@ The structure of mod_lists:
 
     $editor->{mod_lists} = {};
     foreach ( keys %$paths ) {
+	my $spool = SVN::Pool->new_default;
         my $item = $paths->{$_};
 	s/\n/ /g; # XXX: strange edge case
         my $href;
@@ -492,13 +493,26 @@ The structure of mod_lists:
 	$editor->close_directory($root);
 	$editor->close_edit;
     } else {
-        my @mod_list = keys %{$editor->{mod_lists}};
+        my @mod_list = sort keys %{$editor->{mod_lists}};
+	# mark item as directory that we are sure about.
+	# do not use !isdir for deciding the item is _not_ a directory.
+	for my $parent (@mod_list) {
+	    for (@mod_list) {
+		next if $parent eq $_;
+		if (index ("$_/", "$parent/") == 0) {
+		    $editor->{mod_lists}{$parent}{isdir} = 1;
+		    last;
+		}
+	    }
+	}
         if ( ($self->{skip_to} && $self->{skip_to} <= $rev) ||
 	     grep { my $href = $editor->{mod_lists}{$_};
                     !( ( ($href->{action} eq 'A' || $href->{action} eq 'R')
-                         && defined $href->{local_rev}
-                         && $href->{local_rev} != -1
-                         && $href->{source_node_kind} == $SVN::Node::dir)
+                         && ((defined $href->{local_rev}
+			      && $href->{local_rev} != -1
+			      && $href->{source_node_kind} == $SVN::Node::dir)
+			     || ($href->{isdir})
+			    ))
                        || $href->{action} eq 'D' )
                 } @mod_list ) {
 	    my $pool = SVN::Pool->new_default_sub;
@@ -519,17 +533,25 @@ The structure of mod_lists:
 
             $edit->open_root ($self->{headrev});
 
-            foreach (sort @mod_list) {
+            foreach (@mod_list) {
                 my $href = $editor->{mod_lists}{$_};
                 my $action = $href->{action};
 
 		if ($action eq 'D' || $action eq 'R') {
-                    $edit->delete_entry ($_);
+		    # XXX: bad pool usage here, but svn::simple::edit sucks
+                    $edit->delete_entry($_);
                 }
 
+		# can't use a new pool for these, because we need to
+		# keep the parent.  switch to svk dynamic editor when we can
                 if ($action eq 'A' || $action eq 'R') {
-                    $edit->copy_directory ($_, $href->{local_source_path},
-                                           $href->{local_rev});
+		    if (defined $href->{local_rev} && $href->{local_rev} != -1) {
+			$edit->copy_directory ($_, $href->{local_source_path},
+					       $href->{local_rev});
+		    }
+		    else {
+			$edit->add_directory($_);
+		    }
                     $edit->close_directory ($_);
 		}
 	    }
@@ -751,6 +773,20 @@ sub open_directory {
     print "MirrorEditor::open_directory($path)\n" if $debug;
     return undef unless $pb;
 
+
+    if ( ($self->_in_modified_list($path) || '') eq 'R' ) {
+	# if the path is replaced with history, from outside the
+	# mirror anchor... HATE
+	my $bogus_copy = $self->_is_copy($path) && !defined $self->{mod_lists}{$path}{local_source_path};
+	if ($bogus_copy ) {
+	    $self->_create_new_copied_path ($path);
+	}
+	else {
+	    return $self->add_directory($path, $pb, undef, -1, $pool);
+	}
+    }
+
+
     my $dir_baton = $self->SUPER::open_directory ($path, $pb,
                                                   $self->{mirror}{headrev},
                                                   $pool);
@@ -769,6 +805,12 @@ sub open_file {
     my ($self,$path,$pb,undef,$pool) = @_;
     print "MirrorEditor::open_file($path)\n" if $debug;
     return undef unless $pb;
+    my $action = $self->_in_modified_list ($path);
+    if ($self->_under_latest_copypath ($path)
+	&& $self->_is_under_true_copy ($path)
+	&& !$action) {
+	return undef;
+    }
     $self->{opening} = $path;
     return $self->SUPER::open_file ($path, $pb,
 				    $self->{mirror}{headrev}, $pool);
@@ -848,8 +890,11 @@ sub _is_copy {
 sub _get_copy_path_rev {
     my ($self, $path) = @_;
 
-    return ( exists $self->{mod_lists}{$path} ) ?
-        @{$self->{mod_lists}{$path}}{qw/local_source_path local_rev/} : ();
+    return unless exists $self->{mod_lists}{$path};
+    my ($cp_path, $cp_rev) =
+        @{$self->{mod_lists}{$path}}{qw/local_source_path local_rev/};
+    use URI::Escape;
+    return (uri_escape($cp_path), $cp_rev);
 }
 
 # Given a path, return true if it's under the lastest visited copied path.
@@ -1037,7 +1082,7 @@ sub add_directory {
             # 1e.
             $self->delete_entry ($path,
                                  $self->{mirror}{headrev},
-                                 $pb);
+                                 $pb, $pool);
 
             my ($copypath, $copyrev) = $self->_get_copy_path_rev ($path);
             if (defined ($copyrev) && $copyrev >= 0) {
@@ -1165,7 +1210,7 @@ sub add_file {
             $method = 'open_file';
         } elsif ($action eq 'R') {
             # 1d.
-	    $self->delete_entry ($path, $self->{mirror}{headrev}, $pb);
+	    $self->delete_entry ($path, $self->{mirror}{headrev}, $pb, $_[-1]);
 	    # XXX: why should this fail and ignore the following applytext?
 #            return undef;
         }
@@ -1187,7 +1232,10 @@ sub delete_entry {
     my ($self, $path, $rev, $pb, $pool) = @_;
     print "MirrorEditor::delete_entry($path, $rev)\n" if $debug;
     return unless $pb;
-
+    if ($self->_under_latest_copypath($path)) {
+	my $action = $self->_in_modified_list($path) || '';
+	return unless $action eq 'D' || $action eq 'R';
+    }
     $self->SUPER::delete_entry ($path, $self->{mirror}{headrev},
 				$pb, $pool);
 }
